import json
import os.path
from typing import List, Tuple

import numpy as np
import pandas as pd
import scanpy as sc
import torch
from anndata import AnnData
from scipy.special import softmax

from cell_type.common import record_prediction_run_metadata, record_predictions, extract_gene_names, \
    extract_raw_counts_adata, build_prediction_column_name

# Model artifact file names
PYTORCH_MODEL_FILE_NAME = 'model.pt'
FEATURES_FILE_NAME = 'features.csv'
INDEX_TO_LABEL_FILE_NAME = 'idx2label.json'


class LogisticRegression(torch.nn.Module):
    def __init__(self, input_dim, output_dim):
        super(LogisticRegression, self).__init__()
        self.linear = torch.nn.Linear(input_dim, output_dim)

    def forward(self, x):
        outputs = torch.sigmoid(self.linear(x))
        return outputs


def annotate(query_dataset: AnnData, output_h5ad_path, model_artifacts_path,
             annotation_prefix='cell_type', counts_layer=None, gene_column_name=None,
             organism=None, classifier='default'):

    # model artifacts for each classifier level are store in artifact subdir of the same name
    model_artifacts_path = os.path.join(model_artifacts_path, classifier)

    ordered_labels = _load_labels(model_artifacts_path)

    features_used_in_training = _load_features(model_artifacts_path)

    input_dim = len(features_used_in_training)
    output_dim = len(ordered_labels)

    model = _load_model(input_dim, model_artifacts_path, output_dim)

    common_training_features_idx, common_query_features_idx = \
        _compute_common_features(features_used_in_training, query_dataset, gene_column_name, organism)

    weights = model.linear.weight.cpu().detach().numpy().T
    bias = model.linear.bias.cpu().detach().numpy().T

    # normalize raw counts and store in the normalized_raw_counts layer
    x = _normalize_raw_counts(counts_layer, query_dataset)

    x_masked = x[:, common_query_features_idx]
    weights_masked = weights[common_training_features_idx, :]

    probs = softmax(x_masked * weights_masked + bias, 1)
    pred_label_idxs = np.argmax(probs, 1)
    pred_label_names = [ordered_labels[x] for x in pred_label_idxs]

    record_predictions(query_dataset, annotation_prefix, pred_label_names)

    # FIXME(atolopko): pass in source URL for MLflow model instead of model_artifacts_path.
    record_prediction_run_metadata(query_dataset, annotation_prefix, model_artifacts_path)

    uncertainty_col_name = build_prediction_column_name(annotation_prefix, suffix="uncertainty")
    query_dataset.obs[uncertainty_col_name] = 1 - probs.max(1)

    query_dataset.write_h5ad(output_h5ad_path)


def _normalize_raw_counts(counts_layer, query_dataset):
    query_dataset.layers['normalized_raw_counts'] =\
        extract_raw_counts_adata(query_dataset, raw_counts_layer_name=counts_layer)
    sc.pp.normalize_total(query_dataset, layer='normalized_raw_counts', target_sum=1e4)
    sc.pp.log1p(query_dataset, layer='normalized_raw_counts')
    return query_dataset.layers['normalized_raw_counts']


def _compute_common_features(features_in_training_dataset: List[str], query_dataset: AnnData,
                             gene_column_name: str, organism: str) -> Tuple[List[int], List[int]]:
    features_in_query_dataset = extract_gene_names(query_dataset, gene_column_name, organism)
    common_features = set(features_in_query_dataset[features_in_query_dataset.isin(features_in_training_dataset)])

    common_training_features_idx = [i
                                    for i, x in enumerate(features_in_training_dataset)
                                    if x in common_features]

    # Note: the ordering of features in `common_query_features_idx` must match the ordering of the features in
    # `common_training_features_idx`
    features_in_query_dataset_df = pd.DataFrame(features_in_query_dataset)
    features_in_query_dataset_df['query_index'] = features_in_query_dataset.reset_index().index
    common_query_features_idx = \
        pd.merge(pd.DataFrame(features_in_training_dataset, columns=[gene_column_name]),
                 features_in_query_dataset_df,
                 on=gene_column_name)['query_index'].to_list()

    return common_training_features_idx, common_query_features_idx


def _load_labels(model_artifacts_path) -> List[str]:
    with open(os.path.join(model_artifacts_path, INDEX_TO_LABEL_FILE_NAME), 'rb') as f:
        idx2label = json.load(f)
    # TODO: we could just be provided the order labels, rather than having to build it from this dict
    return [idx2label[str(i)] for i in range(len(idx2label))]


def _load_features(model_artifacts_path):
    features_df = pd.read_csv(os.path.join(model_artifacts_path, FEATURES_FILE_NAME))
    features_used_in_training = features_df['feature_name'].values
    return features_used_in_training


def _load_model(input_dim, model_artifacts_path, output_dim):
    model = LogisticRegression(input_dim, output_dim).to('cpu')
    state_dict = torch.load(os.path.join(model_artifacts_path, PYTORCH_MODEL_FILE_NAME),
                            map_location=torch.device('cpu'))
    model.load_state_dict(state_dict)
    model.eval()
    model = model.to('cpu')
    return model
