import json
import os
import re
import shutil
import sys
from datetime import datetime

import mlflow


def init_mlflow_output_dir(mlflow_model_path):
    os.makedirs(mlflow_model_path, exist_ok=True)

    return mlflow_model_path


def build_mlflow_distributable(distributable_file_base_name):
    mlflow_archive = shutil.make_archive(distributable_file_base_name, "zip", distributable_file_base_name)
    print(f"Packaged MLflow model into distributable file '{mlflow_archive}'")


def package_mlflow_model(model_artifacts_path, model_output_path, model_loader_module, conda_requirements_file):
    with open(conda_requirements_file, 'rt') as reqs_file:
        conda_reqs = json.load(reqs_file)

    print(f"Packaging MLflow model using artifacts from '{model_artifacts_path}'")
    mlflow.pyfunc.save_model(path=model_output_path,
                             # registered_model_name='cellxgene_cell_type_annotation',
                             loader_module=model_loader_module,
                             code_path=["src/cell_type"],
                             data_path=model_artifacts_path,
                             conda_env={
                                 "name": "mlflow-env",
                                 # The conda-forge channel is from miniforge, and includes binary packages for arm64
                                 # arch (Apple Silicon: M1, etc.); Using the `conda` channel means installing Python
                                 # packages using `pip`, which can force building them from source, and this is
                                 # failure-prone.
                                 "channels": ["conda-forge"],
                                 "dependencies": conda_reqs
                             })

    build_mlflow_distributable(model_output_path)


# TODO: Retrieve model artifacts from dbfs
# TODO: Upload distributable model files to s3://cellxgene-annotation-public/...
if __name__ == '__main__':
    mlflow_model_output_dir = init_mlflow_output_dir(sys.argv[1])
    local_model_artifacts_base_path = sys.argv[2]
    packaging_timestamp = sys.argv[3] if len(sys.argv) > 3 else datetime.utcnow().strftime("%Y%m%d%H%M%S")

    package_mlflow_model(model_artifacts_path=os.path.join(local_model_artifacts_base_path, 'hlca'),
                         model_output_path=f"{mlflow_model_output_dir}/hlca_{packaging_timestamp}",
                         model_loader_module='cell_type.scvi_models.mlflow_model_loader',
                         conda_requirements_file='src/cell_type/scvi_models/conda-miniforge-requirements.json')
    package_mlflow_model(model_artifacts_path=os.path.join(local_model_artifacts_base_path, 'immune_ref'),
                         model_output_path=f"{mlflow_model_output_dir}/immune_ref_{packaging_timestamp}",
                         model_loader_module='cell_type.scvi_models.mlflow_model_loader',
                         conda_requirements_file='src/cell_type/scvi_models/conda-miniforge-requirements.json')
    # this model's artifacts are obtained via dev/fetch_lr_model_artifacts.sh, which must be run first
    package_mlflow_model(model_artifacts_path=os.path.join(local_model_artifacts_base_path, 'hlca_lr'),
                         model_output_path=f"{mlflow_model_output_dir}/hlca_lr_{packaging_timestamp}",
                         model_loader_module='cell_type.lr_models.mlflow_model_loader',
                         conda_requirements_file='src/cell_type/lr_models/conda-miniforge-requirements.json')


# def setup_mlflow_experiment() -> str:
#     experiment_name = "scanvi"
#     experiment = mlflow.get_experiment_by_name(experiment_name) or \
#                  mlflow.get_experiment(mlflow.create_experiment(experiment_name))
#     print("Name: {}".format(experiment.name))
#     print("Experiment_id: {}".format(experiment.experiment_id))
#     print("Artifact Location: {}".format(experiment.artifact_location))
#     print("Tags: {}".format(experiment.tags))
#     print("Lifecycle_stage: {}".format(experiment.lifecycle_stage))
#     return experiment.experiment_id
#
#
# if __name__ == '__main__':
#     experiment_id = setup_mlflow_experiment()
#
#     with mlflow.start_run(experiment_id=experiment_id) as run:
#         print(f"MLflow run info: {run.info}")
#



# def mlflow_log_h5ad_artifact(ad: AnnData, file_name: str):
#     with TemporaryDirectory() as tmp_dir:
#         file_path = os.path.join(tmp_dir, file_name)
#         ad.write_h5ad(file_path)
#         mlflow.log_artifact(file_path)

# def build_raw_counts_adata( ad) -> AnnData:
#     if self.layer:
#         X = ad.layers[self.layer]
#         var = ad.var
#     elif ad.raw:
#         X = ad.raw.X
#         var = ad.raw.var
#     else:
#         X = ad.X
#         var = ad.var
#     return AnnData(X=X, obs=ad.obs, var=var)
