from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
import os
import scvi
import scanpy as sc
import numpy as np
import json
import pandas as pd
import xgboost as xgb
from typing import Optional

pd.options.display.max_columns = 50
pd.options.display.max_rows = 200


def build_ref_classifiers(adata, classifier_setup, model_path, eval_verbosity: Optional[int] = 1, gpu: Optional[bool] = True):
    """
    This function builds xgboost classifiers on a reference embedding for a designated number of 
    adata.obs columns. Classifier .xgb files and a model_info.json file is written to the `model_path`
    directory. Model evaluation is printed to stdout.
    
    Inputs:
        * `adata`: The AnnData object that was used to train the reference model
        * `classifier_setup`: A dictionary with classifier names (str) to .obs keys 
                              (str) describing the labels for which classifiers should 
                              be built (see Example below)
        * `model_path`: The reference model directory where the classifiers will also be stored
        * `eval_verbosity`: The verbosity level for evaluation of the classifier from the range [0;2].
        * `gpu`: Boolean indicating whether a gpu is available for classifier training
        
        
    Example:
    ```
    In[1]: adata
    
    Out[1]: 
    AnnData object with n_obs × n_vars = 700 × 765
    obs: 'ann_finest_level', 'ann_level_1'

    In [2]: classifier_setup = {
    ...: 'default': 'ann_finest_level',
    ...: 'coarse': 'ann_level_1'
    ...: }
    
    In [3]: ls /path/to/model
    model_params.pt*
    
    In [4]: build_ref_classifiers(adata, classifier_setup, 'path/to/model', eval_verbosity=1, gpu=True)

    In [5]: ls /path/to/model
    classifier_ann_finest_level.xgb*    model_info.json*
    classifier_ann_level_1.xgb*         model_params.pt* 
    ```
    
    """

    # Check inputs
    if not isinstance(eval_verbosity, int):
        raise TypeError('`eval_verbosity` should be an integer between 0 and 2.')
        
    if eval_verbosity < 0 or eval_verbosity > 2:
        raise ValueError('`eval_verbosity` should be an integer between 0 and 2.')

    #classifier names
    if not isinstance(classifier_setup, dict):
        raise TypeError('Please pass a classifier setup dictionary to `classifier_setup`.')
        
    if 'default' not in classifier_setup.keys():
       raise ValueError('`classifier_setup` should contain one classifier named `default`' 
                        ', which will be the default classifier.')
    
    #Note: Currently checking label_keys in build_classifier function
    
    
    # Load ref model
    reference_model = scvi.model.SCANVI.load(model_path, adata=adata)
    
    # Set up classifier dictionary
    classifiers = dict()
    
    for cname, lab in classifier_setup.items():
        filename = 'classifier_'+lab+'.xgb'
        
        print(f'Building classifier for {lab}...')
        xgb_model = build_classifier(
            adata=adata,
            label_key=lab,
            reference_model=reference_model,
            eval_verbosity=eval_verbosity,
            gpu=gpu
        )
        
        # Save classifier
        xgb_model.save_model(os.path.join(model_path, filename))
        
        # Store classifier info
        classifiers[cname] = filename
        

    # Store model_info.json file
    model_params = get_model_params(reference_model)
    model_info = {'model_params':model_params, 'classifier_info':classifiers}

    # Write file
    json_string = json.dumps(model_info, indent=4)
    with open(model_path+'/model_info.json', 'w') as f:
        f.write(json_string)

    
def get_model_params(ref_model):
    """
    Get model parameters from loaded reference model
    """
    
    model_params = dict()

    batch_key = ref_model.adata_manager.registry['setup_args']['batch_key']
    ct_key = ref_model.adata_manager.registry['setup_args']['labels_key']
    unlabeled_category = ref_model.adata_manager.registry['setup_args']['unlabeled_category']
    
    model_params['batch_key'] = batch_key
    model_params['ct_key'] = ct_key
    model_params['unlabeled_category'] = unlabeled_category
    
    return model_params


def build_classifier(adata, label_key, reference_model, eval_verbosity: Optional[int] = 1, gpu=True):
    
    if label_key not in adata.obs:
        raise ValueError(f'{label_key} is not in the `adata` object passed!')
        
    #TODO: check if adata.obs[label_key].dtype != 'category'
    
    # Adata prep
    X_train, X_test, y_train, y_test = setup_anndata(adata, reference_model, label_key)
    #Note: Do we need a new train-test split for each classifier?
    
    # Model training
    xgb_mod = train_xgb_model(X_train, y_train, reference_model, label_key, gpu=gpu)
    
    # Model eval
    if eval_verbosity != 0:
        cr_df = get_model_eval(xgb_mod, X_test, y_test)
        
        if eval_verbosity == 2:
            print(cr_df)

        else:
            overall_accuracy = cr_df['support']['accuracy']
            low_prec_key = cr_df.precision.idxmin()
            low_prec_val = cr_df.precision.min()
            low_rec_key = cr_df.recall.idxmin()
            low_rec_val = cr_df.recall.min()
            low_f1_key = cr_df['f1-score'].idxmin()
            low_f1_val = cr_df['f1-score'].min()

            print('')
            print(f'Summary stats for {label_key} model:')
            print(f'Overall accuracy: {overall_accuracy}')
            print(f'Min. precision: {low_prec_key}: {low_prec_val}')
            print(f'Min. Recall: {low_rec_key}: {low_rec_val}')
            print(f'Min. F1-score: {low_f1_key}: {low_f1_val}')
            print('')
            
    return xgb_mod
            
        
    
    
# Set up training dataframe
def classif_df(adata, trained_model, label_col):
    """
    Note: adata should be the training data for the reference model
    
    """
    
    emb_data = trained_model.get_latent_representation(adata) # Don't actually need adata here at all if model is trained!
    
    df = pd.DataFrame(
        data = emb_data,
        index = adata.obs_names
    )
    
    df['cell_type'] = adata.obs[label_col]
    
    return df
    
    
def setup_anndata(adata, ref_model, label_key):
    
    train_data = classif_df(adata, ref_model, label_col=label_key)

    X_train, X_test, y_train, y_test = train_test_split(train_data.drop(columns='cell_type'), train_data['cell_type'], test_size=0.2, random_state=42, stratify=adata.obs[label_key])
    
    return X_train, X_test, y_train, y_test
    
    
def train_xgb_model(X_train, y_train, ref_model, label_col, gpu=True):

    if gpu:
        tree_method = 'gpu_hist'
    else:
        tree_method = 'hist'
            
    xgbc_gpu = xgb.XGBClassifier(tree_method = tree_method, objective = 'multi:softprob')
    xgbc_gpu.fit(X_train, y_train)
    
    return xgbc_gpu


def get_model_eval(xgb_model, X_test, y_test):
    # Further eval ideas:
    # - uncertainty for correct vs incorrect predictions
    # Note: get uncertainty via: probs = xgb_model.predict_proba(X_test)
    
    preds = xgb_model.predict(X_test)
    
    cr = classification_report(y_test, preds, output_dict=True)
    cr_df = pd.DataFrame(cr).transpose()

    return cr_df


    