import json
import os

import numpy as np
import pandas as pd
import scanpy as sc
import xgboost as xgb
from anndata import AnnData
from pandas import DataFrame

from cell_type.common import record_prediction_run_metadata, record_predictions, build_prediction_column_name
from cell_type.scvi_models.prepare import prep_query_data


#Note: you now need to specify a classifier filename
# - Added a classifier info dictionary with options
# - classifier options always available are: 'default'
# - further options can be levels=['1','2','3'], or categories like 'coarse', 'fine'
# Example from HLCA:
#
# classifiers = {
#    '1': 'classifier_ann_level_1.xgb',
#    '2': 'classifier_ann_level_2.xgb',
#    '3': 'classifier_ann_level_3.xgb',
#    'default': 'classifier_ann_finest_level.xgb',
#    'coarse': 'classifier_ann_coarse_for_GWAS_and_modeling.xgb'
#}


def annotate(query_dataset, output_h5ad_path, model_artifacts_path, classifier='default',
             annotation_prefix='cell_type', counts_layer=None, gene_column_name=None, use_gpu: bool = False,
             train_param_overrides=None, organism='Homo sapiens', max_epochs=500, check_val_every_n_epoch=1):
    if train_param_overrides is None:
        train_param_overrides = {}

    # Load model info file:
    with open(os.path.join(model_artifacts_path, 'model_info.json')) as f:
        model_info = json.load(f)

    # Check that classifier should exist:
    if classifier not in model_info['classifier_info'].keys():
        raise ValueError('Please specify a valid classifier model out of:\n'
                         f"{model_info['classifier_info'].keys()}")

    # Specify classifier path
    cell_type_classifier_model_path = os.path.join(model_artifacts_path, model_info['classifier_info'][classifier])
    reference_embedding_model_dir = model_artifacts_path

    # Get model parameterisation:
    ct_key = model_info['model_params']['ct_key']
    batch_key = model_info['model_params']['batch_key']
    unlabeled_category = model_info['model_params']['unlabeled_category']

    # TODO(atolopko): move the above into a separate read_and_test params function

    # Retrieve (possibly remote) model file and cache locally. Then place the model file in a temp dir with the
    # specific 'model.pt' file name in order to appease scArches' model loading logic, which expects the directory of
    # the model, not the full path to the model.pt file. While scArches appears to have its own support for caching
    # remote files, it doesn't seem to work.

    query_dataset_prepped = prep_query_data(query_dataset,
                                            reference_embedding_model_dir,
                                            counts_layer=counts_layer,
                                            gene_column_name=gene_column_name,
                                            dataset_name=str(query_dataset.filename),
                                            organism=organism,
                                            ct_key=ct_key,
                                            batch_key=batch_key,
                                            unlabeled_category=unlabeled_category
                                            )

    query_dataset_reference_embedding = map_to_ref(query_dataset_prepped,
                                                   reference_embedding_model_dir,
                                                   max_epochs=max_epochs,
                                                   use_gpu=use_gpu,
                                                   check_val_every_n_epoch=check_val_every_n_epoch,
                                                   train_kwargs_overrides=train_param_overrides)
    query_dataset.obsm[annotation_prefix] = query_dataset_reference_embedding

    # FUTURE WORK: Put a for loop around this for multiple classifiers per reference
    # - would need to add classifier name to annotation_column_name
    obs_predictions = predict(query_dataset_prepped,
                              query_dataset_reference_embedding,
                              cell_type_classifier_model_path,
                              annotation_prefix=annotation_prefix,
                              use_gpu=use_gpu)
    query_dataset.obs = pd.concat([query_dataset.obs, obs_predictions], axis=1)

    generate_query_dataset_umap(query_dataset, annotation_prefix)

    # FIXME(atolopko): pass in source URL for MLflow model instead of model_artifacts_path.
    record_prediction_run_metadata(query_dataset, annotation_prefix, model_artifacts_path)

    query_dataset.write_h5ad(output_h5ad_path)


def predict(
        query_dataset: AnnData,
        query_dataset_embedding,
        cell_type_classifier_model_path,
        annotation_prefix: str,
        use_gpu: bool = False
) -> DataFrame:
    """
    Returns `obs` DataFrame with prediction columns appended
    """
    tree_method = 'gpu_hist' if use_gpu else 'hist'
    cell_type_classifier_model = xgb.XGBClassifier(tree_method=tree_method, objective='multi:softprob')
    cell_type_classifier_model.load_model(fname=cell_type_classifier_model_path)

    project_labels(query_dataset, cell_type_classifier_model, query_dataset_embedding,
                   annotation_prefix=annotation_prefix)

    return query_dataset.obs


# def update(:
#     input_data.obs[predicted_label_name] = predictions
#     # TODO: obj.obs[f'{model_key}_conf_score'] = conf_scores
#     input_data.obsm[f"X_{model_key}"] = model.get_latent_representation(input_data)
#
# def validate(:
#     """
#     calculate validation metrics:
#         F1 - score
#     """
#     f1_score = metrics.f1_score(input_data.obs[label_name],
#                                      input_data.obs[predicted_label_name],
#                                      average='micro')
#     mlflow.log_metric("f1", f1_score)
#     print(f"f1={f1_score}")

# def save( output_h5ad_file: Union[pathlib.Path, str]):
#     input_data.write_h5ad(output_h5ad_file)
#     print(f"saved predictions to {output_h5ad_file}")

def install_local_model(reference_embedding_model_locator, tmp_ref_emb_model_dir):
    with reference_embedding_model_locator.open() as f_in, \
            open(os.path.join(tmp_ref_emb_model_dir, 'model.pt'), "wb") as f_out:
        f_out.write(f_in.read())


def generate_query_dataset_umap(query_dataset, annotation_prefix):
    neighbors_key = f"{annotation_prefix}_neighbors"
    sc.pp.neighbors(query_dataset, use_rep=annotation_prefix, key_added=neighbors_key)
    query_dataset_with_umap = sc.tl.umap(query_dataset, neighbors_key=neighbors_key, copy=True)
    query_dataset.obsm[f"X_{annotation_prefix}_umap"] = query_dataset_with_umap.obsm['X_umap']




# Parameters
early_stopping_kwargs_scarches = {
    'enable_progress_bar': True,  # Fix for when using torch nightly build
    'early_stopping': True,
    'early_stopping_monitor': 'elbo_train',
    'early_stopping_patience': 10,
    'early_stopping_min_delta': 0.001,
}

plan_kwargs = {
    "reduce_lr_on_plateau": True,
    "lr_patience": 8,
    "lr_factor": 0.1,
    "weight_decay": 0,
}


def map_to_ref(query_dataset,
               model_path,
               max_epochs=500,
               use_gpu=False,
               train_kwargs_overrides={},
               plan_kwargs_overrides={},
               check_val_every_n_epoch=1,
               ) -> np.ndarray:
    """
    A function to map the query data to the reference atlas

    Input:
        * query_dataset: An AnnData object that has been prepped by the `prep_query_data` function
        * model_path: The HLCA reference model directory
        * max_epochs: The maximum number of epochs for reference training (Default: 500)
        * use_gpu: Boolean flag whether to use gpu for training (Default: True)
        * train_kwargs: Keyword arguments passed to `scvi.model.SCANVI.train()` for early stopping
        * plan_kwargs: Dictionary of training plan keyword arguments passed to `scvi.model.SCANVI.train()`

    Output:
        A numpy.ndarray containing the embedding coordinates of the query data in the HLCA latent space.
        If `output_model` is set to True, then the trained reference model is also output

    """
    import scvi

    # Load query data into the model
    vae_q = scvi.model.SCANVI.load_query_data(
            query_dataset,
            model_path,
            freeze_dropout=True,
    )

    # Train scArches model for query mapping
    vae_q.train(
            max_epochs=max_epochs,
            check_val_every_n_epoch=check_val_every_n_epoch,
            use_gpu=use_gpu,
            plan_kwargs={**plan_kwargs, **plan_kwargs_overrides},  # Do overrides really go after the
            **({**early_stopping_kwargs_scarches, **train_kwargs_overrides}),
    )

    emb = vae_q.get_latent_representation()

    return emb


def project_labels(query_dataset: AnnData,
                   cell_type_classifier_model: xgb.XGBClassifier,
                   embedding_coords: np.ndarray,
                   annotation_prefix='cell_type',
                   uncertainty_thresh=None  # Note: currently not passed to predict function
                   ):
    """
    A function that projects predicted labels onto the query dataset, along with uncertainty scores.
    Performs in-place update of the adata object, adding columns to the `obs` DataFrame.

    Input:
        * `query_dataset`: The query `AnnData` object which was mapped to the HLCA reference
        * `model_file`: Path to the classification model file
        * `embedding_coords`: A numpy.ndarray of the latent space embedding coordinates of the
                            query data mapped to the HLCA reference. This is output by the
                            `map_to_ref()` function
        * `prediction_key`: Column name in `adata.obs` where to store the predicted labels
        * `uncertainty_thresh`: The uncertainty threshold above which we call a cell 'Unknown'

    Output:
        Nothing is output, the passed anndata is modified inplace

    """
    if (uncertainty_thresh is not None) and (uncertainty_thresh < 0 or uncertainty_thresh > 1):
        raise ValueError(f'`uncertainty_thresh` must be `None` or between 0 and 1.')

    # Set up query data
    q_emb_df = pd.DataFrame(data=embedding_coords, index=query_dataset.obs_names)

    # Predict labels
    pred_labels = cell_type_classifier_model.predict(q_emb_df)
    record_predictions(query_dataset, annotation_prefix, pred_labels)

    # Predict probabilities
    probs = cell_type_classifier_model.predict_proba(q_emb_df)

    # Format probabilities
    df_probs = pd.DataFrame(probs, columns=cell_type_classifier_model.classes_, index=query_dataset.obs_names)

    # TODO: Check this! Does that selected max correspond to the label that was predicted? #Malte: checked this
    uncertainty_col_name = build_prediction_column_name(annotation_prefix, suffix="uncertainty")
    query_dataset.obs[uncertainty_col_name] = 1 - df_probs.max(1)

    # TODO: Update caller to copy this to output AnnData object, if we desired this
    # adata.uns[uncertainty_col_name] = df_probs

    # Note: this is here in case we want to propose a set of values for the user to accept to seed the
    #       manual curation of predicted labels
    if uncertainty_thresh is not None:
        confident_col_name = build_prediction_column_name(annotation_prefix, "confident")
        uncertainty_col_name = build_prediction_column_name(annotation_prefix, "uncertainty")
        prediction_col_name = build_prediction_column_name(annotation_prefix)
        query_dataset.obs[confident_col_name] = [
            val if query_dataset.obs[uncertainty_col_name][i] < uncertainty_thresh
            else 'Unknown' for i, val in enumerate(query_dataset.obs[prediction_col_name])]
