import numpy as np
import pandas as pd
from anndata import AnnData

from cell_type.common import extract_raw_counts_adata, extract_gene_names


def build_query_dataset(adata, counts_layer, gene_column_name, organism):
    raw_counts = extract_raw_counts_adata(adata, counts_layer)
    gene_names = extract_gene_names(adata, gene_column_name, organism)
    return AnnData(X=raw_counts,
                   var=pd.DataFrame(index=gene_names),
                   obs=pd.DataFrame(index=adata.obs_names))

    
def prep_query_data(
        adata,
        model_path,
        counts_layer='counts',
        gene_column_name=None,
        min_counts = 200,
        dataset_name='query_data',
        organism='Homo sapiens',
        ct_key='Harmonised_detailed_type',
        batch_key="Dataset",
        unlabeled_category="unlabeled",
): # prep adata object for ref mapping; output -> prepped query object
    """
    A function to prepare the `AnnData` object for reference mapping to the HCLA

    Note: other reference models may have other requirements, such as:
        * count data stored in a different location (not a layer?)
        * size_factors stored or to be computed
        * feature names no HGNC gene names
        * ct_key, batch_key, and unlabeled_category are ref-model-specific and could be stored with the model.
          could dig into the model load code to get these directory from the ref model

    Input:
        * `adata`: A query AnnData object with count data stored in adata.layers
        * `model_path`: The HLCA reference model directory
        * `counts_layer`: The name of the layer containing count data. If None, uses raw.X, if available, otherwise X
        * `min_counts`: The minimum number of counts per cell allowed for query data to start ref mapping
        * `dataset_name`: (Optional) The name of the dataset
        * `organism`: The organism from which the data dervies. Only 'Homo sapiens' and 'Mus musculus' are supported.
        * `ct_key`: Obs column that contains the cell type labels used for integration in the reference
        * `batch_key`: Obs column that contains the batch labels in the reference
        * `unlabeled_category`: Category the reference model recognizes as an unknown label
        * 

    Output:
        A prepared `AnnData` object that can be loaded into `map_to_ref` for reference mapping
    """
    import scanpy as sc
    import scvi

    # Prep query anndata
    # Checks to implement:
    # - test type of input
    # - var names are gene names -> how to test? Just check some examples?
    # - layer contains non-negative integers
    if not isinstance(adata, sc.AnnData):
        raise TypeError(f'`adata` should contain an AnnData object!')

    if not isinstance(dataset_name, str):
        raise TypeError('`dataset_name` should be a string!')

    if organism not in ['Homo sapiens', 'Mus musculus']:
        raise ValueError('`organism` should be one of `Homo sapiens` or `Mus musculus`.')
    
    adata_query = build_query_dataset(adata, counts_layer, gene_column_name, organism)

    if np.any(adata_query.X.sum(1) < min_counts):
        raise ValueError(f'There are cells with fewer than {min_counts} counts in the data.\n'
                         'Please filter your data with `sc.pp.filter_cells()` prior to reference mapping.')

    #TODO: check if batch_key or ct_key already exist in the query data
    adata_query.obs[batch_key] = dataset_name
    adata_query.obs[ct_key] = unlabeled_category

    # Prepare AnnData object to match to scANVI query data
    scvi.model.SCANVI.prepare_query_anndata(adata_query, model_path)

    # Note: can adapt this now that prepare_query_anndata supports layers (see: https://github.com/scverse/scvi-tools/pull/1575)
    adata_query.layers['counts'] = adata_query.X

    return adata_query


